from ..core import glinter
from ..util import set_executable
from .cmake import section_check
from ros_introspect.package import DependencyType
import re

MAINPAGE_S = r'/\*\*\s+\\mainpage\s+\\htmlinclude manifest.html\s+\\b %s\s+<!--\s+' + \
             r'Provide an overview of your package.\s+-->\s+-->\s+[^\*]*\*/'


@glinter
def check_dynamic_reconfigure(package):
    if not package.dynamic_reconfig:
        return
    pkg_list = {'dynamic_reconfigure'}

    package.package_xml.add_dependencies({DependencyType.BUILD: pkg_list, DependencyType.RUN: pkg_list})

    cmake = package.cmake
    section_check(cmake, [str(cfg.rel_fn) for cfg in package.dynamic_reconfig],
                  'generate_dynamic_reconfigure_options', '')
    section_check(cmake, pkg_list, 'find_package', 'COMPONENTS')

    for cfg in package.dynamic_reconfig:
        set_executable(cfg.full_path, True)


@glinter
def remove_useless_files(package):
    mainpage_pattern = re.compile(MAINPAGE_S % package.name)
    for doc in package.documentation[:]:
        if doc.full_path.name != 'mainpage.dox':
            continue

        s = open(doc.full_path).read()
        if mainpage_pattern.match(s):
            doc.full_path.unlink()
            package.documentation.remove(doc)


@glinter
def misc_xml_formatting(package):
    package.package_xml.force_regeneration()
    for config in package.plugin_xml:
        config.force_regeneration()



from colorama import Fore, Back, Style, init
import shutil
import sys

init()


if sys.stdout.encoding == 'utf-8':
    TL = '╭'
    TR = '╮'
    BL = '╘'
    BR = '╛'
    TC = '─'
    BC = '═'
    VC = '│'
else:
    TL = TR = BL = BR = '+'
    TC = BC = '-'
    VC = '|'


def color_diff(diff, buffer_size=5):
    # Buffer size is the number of lines before and after diffs to print
    buffer = []
    started = False

    for line in diff:
        color = ''
        if line.startswith('+'):
            color = Fore.GREEN
        elif line.startswith('-'):
            color = Fore.RED
        elif line.startswith('^'):
            color = Fore.BLUE
        elif line.startswith('?'):
            continue

        if color:
            if buffer:
                if len(buffer) > buffer_size:
                    yield '...'
                yield from buffer[-buffer_size:]
            started = True
            yield color + line + Fore.RESET
            buffer = []
        else:
            if started and len(buffer) >= buffer_size:
                yield from buffer[:buffer_size]
                buffer = []
                started = False
            buffer.append(line)
    if buffer:
        yield from buffer[-buffer_size:]
        yield '...'


def color_header(s, fore='WHITE', back='BLUE'):
    ts = shutil.get_terminal_size()
    header = ''
    header += getattr(Fore, fore) + getattr(Back, back)
    header += TL + (TC * (ts.columns - 2)) + TR + '\n'
    header += VC + ' ' + Style.BRIGHT
    header += s.ljust(ts.columns - 3, ' ') + Style.NORMAL + VC + '\n'
    header += BL + (BC * (ts.columns - 2)) + BR
    header += Style.RESET_ALL
    return header


def color_text(s, fore='YELLOW', bright=False):
    return (Style.BRIGHT if bright else '') + getattr(Fore, fore) + s + Style.RESET_ALL



from ..core import glinter
from ..util import get_ignore_data
from ..config import get_config
from ros_introspect.package import DependencyType
from ros_introspect.components.package_xml import count_trailing_spaces, get_chunks, get_sort_key, PEOPLE_TAGS


@glinter
def check_manifest_dependencies(package, config=None):
    if config is None:
        config = get_config()  # pragma: no cover
    prefer_depend_tag = config.get('prefer_depend_tag', False)

    dep_dict = {}
    for dt in DependencyType:
        dep_dict[dt] = package.get_dependencies(dt)

    package.package_xml.add_dependencies(dep_dict, prefer_depend_tag)

    # Special handling for interface dependencies
    # because not all run deps are build deps
    if not package.get_ros_interfaces():
        return
    if package.ros_version == 1:
        build_dep = 'message_generation'
        run_dep = 'message_runtime'
        export = 'message_runtime'
        export_tag = 'build_export_depend'
    else:
        build_dep = 'rosidl_default_generators'
        run_dep = 'rosidl_default_runtime'
        export = 'rosidl_interface_packages'
        export_tag = 'member_of_group'

    if package.package_xml.xml_format == 1:
        pairs = [('build_depend', build_dep),
                 ('run_depend', run_dep)]
    else:
        pairs = [('build_depend', build_dep),
                 (export_tag, export),
                 ('exec_depend', run_dep)]
        package.package_xml.remove_dependencies('depend', [build_dep, run_dep])
    for tag, msg_pkg in pairs:
        existing = package.package_xml.get_packages_by_tag(tag)
        if msg_pkg not in existing:
            package.package_xml.insert_new_packages(tag, [msg_pkg])


@glinter
def enforce_manifest_ordering(package, alphabetize=True):
    root = package.package_xml.root
    chunks = get_chunks(root.childNodes)

    new_children = []

    for a, b in sorted(chunks, key=lambda d: get_sort_key(d[0], alphabetize, package.package_xml.xml_format)):
        new_children += b

    if root.childNodes != new_children:
        package.package_xml.changed = True
        root.childNodes = new_children


@glinter
def remove_empty_export_tag(package):

    def has_element_child(node):
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                return True
        return False

    manifest = package.package_xml
    exports = manifest.root.getElementsByTagName('export')
    if len(exports) == 0:
        return False
    for export in exports:
        if not has_element_child(export):
            manifest.remove_element(export)
            return True


@glinter
def enforce_manifest_tabbing(package):
    def enforce_tabbing_helper(manifest, node, tabs=1):
        ideal_length = manifest.std_tab * tabs
        prev_was_node = True
        insert_before_list = []
        if not node:
            return
        changed = False
        for c in node.childNodes:
            if c.nodeType == c.TEXT_NODE:
                prev_was_node = False
                if c == node.childNodes[-1]:
                    continue

                if '\n' not in c.data:
                    c.data = '\n' + c.data
                    changed = True
                spaces = count_trailing_spaces(c.data)
                if spaces != ideal_length:
                    last_nl = c.data.rindex('\n')
                    c.data = c.data[: last_nl + 1] + (' ' * ideal_length)
                    changed = True
            else:
                if prev_was_node:
                    changed = True
                    insert_before_list.append(c)

                prev_was_node = True
                enforce_tabbing_helper(manifest, c, tabs + 1)

        for c in insert_before_list:
            node.insertBefore(manifest.create_new_tab_element(tabs), c)

        manifest.changed = manifest.changed or changed

        if len(node.childNodes) == 0:
            return
        last = node.childNodes[-1]
        if last.nodeType != last.TEXT_NODE:
            node.appendChild(manifest.create_new_tab_element(tabs - 1))
            manifest.changed = True

    enforce_tabbing_helper(package.package_xml, package.package_xml.root)


@glinter
def remove_empty_manifest_lines(package):
    def remove_empty_lines_helper(node):
        changed = False
        for child in node.childNodes:
            if child.nodeType == child.TEXT_NODE:
                while '\n\n\n' in child.data:
                    child.data = child.data.replace('\n\n\n', '\n\n')
                    changed = True
            else:
                changed = remove_empty_lines_helper(child) or changed
        return changed

    if remove_empty_lines_helper(package.package_xml.root):
        package.package_xml.changed = True


def cleanup_text_elements(node):
    new_children = []
    changed = False

    for child in node.childNodes:
        if child.nodeType == child.TEXT_NODE and len(new_children) and new_children[-1].nodeType == child.TEXT_NODE:
            changed = True
            new_children[-1].data += child.data
        elif child.nodeType == child.TEXT_NODE and child.data == '':
            continue
        else:
            new_children.append(child)

    node.childNodes = new_children
    return changed


def replace_text_node_contents(node, ignorables):
    changed = False
    removable = []
    for i, c in enumerate(node.childNodes):
        if c.nodeType == c.TEXT_NODE:
            continue
        elif c.nodeType == c.COMMENT_NODE:
            short = c.data.strip()
            if short in ignorables:
                removable.append(i)
                changed = True
                continue
        else:
            changed = replace_text_node_contents(c, ignorables) or changed
    for node_index in reversed(removable):  # backwards not to affect earlier indices
        if node_index > 0:
            before = node.childNodes[node_index - 1]
            if before.nodeType == c.TEXT_NODE:
                trailing = count_trailing_spaces(before.data)
                before.data = before.data[:-trailing]

        if node_index < len(node.childNodes) - 1:
            after = node.childNodes[node_index + 1]
            if after.nodeType == c.TEXT_NODE:
                while len(after.data) and after.data[0] == ' ':
                    after.data = after.data[1:]
                if len(after.data) and after.data[0] == '\n':
                    after.data = after.data[1:]

        node.childNodes.remove(node.childNodes[node_index])
    changed = cleanup_text_elements(node) or changed
    return changed


@glinter
def remove_boilerplate_manifest_comments(package):
    ignorables = get_ignore_data('package', {'package': package.name}, add_newline=False)
    changed = replace_text_node_contents(package.package_xml.root, ignorables)
    if changed:
        package.package_xml.changed = changed
        remove_empty_manifest_lines(package)


def replace_package_set(package_xml, source_tags, new_tag):
    """Replace all the elements with tags in source_tags with new elements with new_tag."""
    intersection = None
    for tag in source_tags:
        pkgs = set(package_xml.get_packages_by_tag(tag))
        if intersection is None:
            intersection = pkgs
        else:
            intersection = intersection.intersection(pkgs)
    for tag in source_tags:
        package_xml.remove_dependencies(tag, intersection)
    package_xml.insert_new_packages(new_tag, intersection)


@glinter
def greedy_depend_tag(package):
    if package.package_xml.xml_format == 1:
        return
    replace_package_set(package.package_xml, ['build_depend', 'build_export_depend', 'exec_depend'], 'depend')


@glinter
def update_people(package, config=None):
    if config is None:
        config = get_config()  # pragma: no cover
    for d in config.get('replace_rules', []):
        target_name = d['to']['name']
        target_email = d['to']['email']
        search_name = d['from'].get('name')
        search_email = d['from'].get('email')

        for el in package.package_xml.get_elements_by_tags(PEOPLE_TAGS):
            name = el.childNodes[0].nodeValue
            email = el.getAttribute('email') if el.hasAttribute('email') else ''
            if (search_name is None or name == search_name) and (search_email is None or email == search_email):
                el.childNodes[0].nodeValue = target_name
                if target_email:
                    el.setAttribute('email', target_email)
                package.package_xml.changed = True


@glinter
def update_license(package, config=None):
    if config is None:
        config = get_config()  # pragma: no cover
    if 'default_license' not in config or 'TODO' not in package.package_xml.get_license():
        return

    package.package_xml.set_license(config['default_license'])



import collections
from enum import IntEnum
import fnmatch
import os.path

from stylish_cmake_parser import Command

from ros_introspect.components.setup_py import create_setup_py

from ..core import glinter
from ..cmake_ordering import insert_in_order
from ..python_types import get_python_usage, PythonUsage
from .cmake import check_complex_section, section_check, get_multiword_section, install_cmake_dependencies


class InstallType(IntEnum):
    EXECUTABLE = 1
    LIBRARY = 2
    HEADERS = 3
    SHARE = 4
    PYTHON = 5


# The destination map is dictionary from destinations to CMake section names
InstallConfig = collections.namedtuple('InstallConfig', ['section_name', 'destination_map'])


INSTALL_CONFIGS = {
    # ROS 1
    1: {
        InstallType.EXECUTABLE: InstallConfig('TARGETS',
                                              {'${CATKIN_PACKAGE_BIN_DESTINATION}': ['RUNTIME DESTINATION']}),
        InstallType.LIBRARY: InstallConfig('TARGETS',
                                           {'${CATKIN_PACKAGE_LIB_DESTINATION}': ['ARCHIVE DESTINATION',
                                                                                  'LIBRARY DESTINATION'],
                                            '${CATKIN_GLOBAL_BIN_DESTINATION}': ['RUNTIME DESTINATION']}),
        InstallType.HEADERS: InstallConfig('FILES', {'${CATKIN_PACKAGE_INCLUDE_DESTINATION}': ['DESTINATION']}),
        InstallType.SHARE: InstallConfig('FILES', {'${CATKIN_PACKAGE_SHARE_DESTINATION}': ['DESTINATION']}),
    },
    2: {
        InstallType.EXECUTABLE: InstallConfig('TARGETS', {'lib/${PROJECT_NAME}': ['DESTINATION']}),
        InstallType.LIBRARY: InstallConfig('TARGETS',
                                           {'lib': ['ARCHIVE DESTINATION', 'LIBRARY DESTINATION'],
                                            'bin': ['RUNTIME DESTINATION'],
                                            }),
        InstallType.HEADERS: InstallConfig('FILES', {'include': ['DESTINATION']}),
        InstallType.SHARE: InstallConfig('FILES', {'share/${PROJECT_NAME}': ['DESTINATION']}),
        InstallType.PYTHON: InstallConfig('PROGRAMS', {'lib/${PROJECT_NAME}': ['DESTINATION']}),
    }
}


def get_install_type(destination, section_name=None):
    """For a given catkin destination, return the matching install type."""
    for config_map in INSTALL_CONFIGS.values():
        for name, (kw, destination_map) in config_map.items():
            if destination not in destination_map:
                continue
            if kw == section_name or (kw == 'FILES' and section_name == 'DIRECTORY'):
                return name


def directory_matches(cmd, subfolder=''):
    the_dir = cmd.get_section('DIRECTORY')
    return the_dir and subfolder and subfolder in the_dir.values


def get_destinations(cmd):
    for section in cmd.get_real_sections():
        if isinstance(section.name, str) and section.name.endswith('DESTINATION'):
            destination = section.values[0]
            if '}' in destination:
                ci = destination.rindex('}') + 1
                root = destination[:ci]
                subfolder = destination[ci:]
                yield root, subfolder
            else:
                yield destination, ''


def get_install_types(cmd, subfolder=''):
    """For a given CMake command, determine the install type(s) that this command uses.

    If there is a non-empty subfolder, we only return the install types if the command
    installs into the catkin_destination with the given subfolder
    """
    types = set()
    check_subfolder = subfolder is not None and not directory_matches(cmd, subfolder)
    token = cmd.get_tokens(True)[0]

    for root_destination, cmd_subfolder in get_destinations(cmd):
        if check_subfolder:
            cmd_subfolder = cmd_subfolder.lstrip('/')
            if subfolder != cmd_subfolder:
                continue
        type_ = get_install_type(root_destination, token)
        if type_:
            types.add(type_)
    return types


def install_sections(cmd, destination_map, subfolder=''):
    """Ensure that the command has all the appropriate CMake sections with the matching catkin destinations.

    If the subfolder is defined, the subfolder is appended to the catkin destination.
    """
    for destination, section_names in destination_map.items():
        for section_name in section_names:
            if len(subfolder) > 0:
                destination = os.path.join(destination, subfolder)
            elif destination == '${CATKIN_PACKAGE_SHARE_DESTINATION}':
                # ensure unslashed version is not present
                section = cmd.get_section(section_name)
                if section and destination in section.values:
                    section.values.remove(destination)
                    section.mark_changed()
                destination += '/'
            check_complex_section(cmd, section_name, destination)


def remove_install_section(cmd, destination_map):
    empty_sections_to_remove = {}
    for destination, section_names in destination_map.items():
        for section_name in section_names:
            parts = section_name.split()
            if len(parts) == 2:
                empty_sections_to_remove[parts[0]] = destination
    sections = cmd.get_real_sections()
    to_remove = []
    for i, section in enumerate(sections):
        if section.name not in empty_sections_to_remove or len(section.values) != 0:
            continue
        next_section = sections[i + 1]
        dest = empty_sections_to_remove[section.name]
        if next_section.name == 'DESTINATION' and len(next_section.values) == 1 and next_section.values[0] == dest:
            to_remove.append(section)
            to_remove.append(next_section)
    if len(to_remove) > 0:
        for section in to_remove:
            cmd.sections.remove(section)
        cmd.mark_changed()


def get_commands_by_type(cmake, install_type, subfolder=''):
    matches = []
    for cmd in cmake.content_map['install']:
        if install_type in get_install_types(cmd, subfolder):
            matches.append(cmd)
    return matches


def matches_patterns(item, patterns):
    for pattern in patterns:
        if pattern[0] == pattern[-1] and pattern[0] == '"':
            pattern = pattern[1:-1]
        if fnmatch.fnmatch(item, pattern):
            return True


def install_section_check(cmake, items, install_type, ros_version, directory=False, subfolder=''):
    install_config = INSTALL_CONFIGS[ros_version][install_type]

    if directory and install_config.section_name == 'FILES':
        section_name = 'DIRECTORY'
    else:
        section_name = install_config.section_name

    cmds = get_commands_by_type(cmake, install_type, subfolder)
    if len(items) == 0:
        for cmd in cmds:
            if len(get_install_types(cmd, subfolder)) == 1:
                cmake.remove_command(cmd)
            else:
                remove_install_section(cmd, install_config.destination_map)

    cmd = None
    items = [os.path.join(subfolder, item) for item in items]
    for cmd in cmds:
        if directory_matches(cmd, subfolder):
            return
        install_sections(cmd, install_config.destination_map, subfolder)
        section = cmd.get_section(section_name)
        if not section:
            if section_name != 'FILES':
                continue
            section = cmd.get_section('DIRECTORY')
            if not section:
                continue
            pattern = get_multiword_section(cmd, ['FILES_MATCHING', 'PATTERN'])
            nonmatching_items = []
            for item in items:
                if pattern and not matches_patterns(item, pattern.values):
                    nonmatching_items.append(item)
            items = nonmatching_items
        else:
            # We match the section
            if set(section.values) != set(items):
                section.values = [value for value in section.values if value in items]
                section.mark_changed()
            items = [item for item in items if item not in section.values]

    if len(items) == 0:
        return

    if cmd is None:
        cmd = Command('install', parent=cmake)
        cmd.add_section(section_name, items)
        insert_in_order(cmake, cmd)
        install_sections(cmd, install_config.destination_map, subfolder)
    elif section:
        # section = cmd.get_section(section_name)
        section.add_values(items, alpha_order=False)

    if subfolder and len(str(cmd)) > 120 and section_name == 'FILES':
        section = cmd.get_section(section_name)
        section.name = 'DIRECTORY'
        section.values = [subfolder + '/']
        section.mark_changed()


@glinter
def update_cplusplus_installs(package):
    if not package.cmake:
        return
    cmake = package.cmake
    install_section_check(cmake, package.cmake.get_executables(), InstallType.EXECUTABLE, package.ros_version)
    install_section_check(cmake, package.cmake.get_libraries(), InstallType.LIBRARY, package.ros_version)

    has_header_files = any(package.get_source_by_tags('header'))
    if package.name and has_header_files:
        if package.ros_version == 1:
            install_section_check(cmake, ['include/${PROJECT_NAME}/'], InstallType.HEADERS, package.ros_version,
                                  directory=True)
        else:
            if not cmake.content_map['ament_export_include_directories']:
                cmd = Command('ament_export_include_directories', parent=cmake)
                cmd.add_section('', ['include'])
                insert_in_order(cmake, cmd)
            install_section_check(cmake, ['include/'], InstallType.HEADERS, package.ros_version, directory=True)


@glinter
def export_cplusplus_libraries(package):
    if package.ros_version == 1:
        return
    libraries = package.cmake.get_libraries()
    if not libraries:
        return

    export_targets = []
    cmds = get_commands_by_type(package.cmake, InstallType.LIBRARY)

    for library in libraries:
        exporting_cmd = None
        lib_cmds = []

        my_export_name = None
        for cmd in cmds:
            targets = cmd.get_section('TARGETS')
            if not targets or library not in targets.values:
                continue
            lib_cmds.append(cmd)
            section = cmd.get_section('EXPORT')
            if section:
                my_export_name = section.values[0]
                exporting_cmd = cmd

        if package.cmake.content_map['ament_export_targets']:
            my_export_name = package.cmake.content_map['ament_export_targets'][0].first_token()

        if my_export_name is None:
            my_export_name = 'export_' + library

        export_targets.append(my_export_name)

        if not exporting_cmd:
            if lib_cmds:
                exporting_cmd = lib_cmds[0]
            else:
                exporting_cmd = Command('install', parent=package.cmake)
                exporting_cmd.add_section('TARGETS', [library])
                insert_in_order(package.cmake, exporting_cmd)

        check_complex_section(exporting_cmd, 'EXPORT', my_export_name)

        # Sort the sections to put EXPORT right after TARGETS
        sections_by_name = {s.name: s for s in exporting_cmd.get_real_sections()}
        assert exporting_cmd.sections.index(sections_by_name['TARGETS']) == 0
        ex_index = exporting_cmd.sections.index(sections_by_name['EXPORT'])

        # Reorder sections
        initial = exporting_cmd.sections[:1]
        export = exporting_cmd.sections[ex_index:ex_index + 1]
        intermediate = exporting_cmd.sections[1:ex_index]
        tail = exporting_cmd.sections[ex_index+1:]
        exporting_cmd.sections = initial + export + intermediate + tail

    section_check(package.cmake, export_targets, 'ament_export_targets')


def check_cmake_python_buildtype(package):
    if package.build_type == 'ament_cmake':
        acp = 'ament_cmake_python'
        build_tools = package.package_xml.get_packages_by_tag('buildtool_depend')
        if acp not in build_tools:
            package.package_xml.insert_new_packages('buildtool_depend', [acp])
            package.build_type = acp

        install_cmake_dependencies(package, {acp})

    if package.setup_py is None:
        create_setup_py(package)


@glinter
def update_misc_installs(package):
    extra_files_by_folder = collections.defaultdict(list)

    for rel_fn, package_file in sorted(package.components_by_name.items()):
        if package_file.__class__.needs_share_installation():
            extra_files_by_folder[rel_fn.parent].append(str(rel_fn.name))

    if package.cmake:
        existing_install_folders = set()
        for cmd in get_commands_by_type(package.cmake, InstallType.SHARE, None):
            for destination, subfolder in get_destinations(cmd):
                existing_install_folders.add(subfolder.lstrip('/'))

        for folder, files in sorted(extra_files_by_folder.items()):
            subfolder = str(folder)
            if subfolder == '.':
                subfolder = ''

            install_section_check(
                package.cmake, files, InstallType.SHARE, package.ros_version, subfolder=subfolder)

            if subfolder in existing_install_folders:
                existing_install_folders.remove(subfolder)

        for subfolder in existing_install_folders:
            install_section_check(package.cmake, [], InstallType.SHARE, package.ros_version, subfolder=subfolder)

    if get_python_usage(package) != PythonUsage.NONE:
        check_cmake_python_buildtype(package)
        if package.ros_version == 2:
            for folder, files in sorted(extra_files_by_folder.items()):
                package.setup_py.include_data_files(files, folder)


@glinter
def fix_double_directory_installs(package):
    if not package.cmake:
        return
    for cmd in package.cmake.content_map['install']:
        dir_section = cmd.get_section('DIRECTORY')
        dest_sections = cmd.get_sections('DESTINATION')

        if not dir_section or not dest_sections:
            continue
        directory = dir_section.values[0]
        final_slash = directory[-1] == '/'

        for section in dest_sections:
            destination = section.values[0]
            if not final_slash and destination.endswith(directory):
                # Remove double directory and final slash
                section.values[0] = destination[:-len(directory) - 1]
                cmd.mark_changed()



import argparse
import pathlib
from . import get_linters
from ros_introspect import find_packages


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('linters', nargs='*', metavar='linter', default=[])
    parser.add_argument('-f', '--related_files', type=pathlib.Path, nargs='*')

    args = parser.parse_args()
    pkgs = {}

    for related_file in args.related_files:
        for pkg in find_packages(related_file.parent):
            if pkg.root not in pkgs:
                pkgs[pkg.root] = pkg

    ret = 0

    for _, package in sorted(pkgs.items()):
        for name, fne in get_linters().items():
            if args.linters and name not in args.linters:
                continue
            fne(package)
        if package.has_changes():
            package.save()
            ret = -1

    return ret



import os
import importlib_resources as resources
import stat


def set_executable(fn, state):
    existing_permissions = stat.S_IMODE(os.lstat(fn).st_mode)
    if state:
        flags = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    else:
        flags = ~stat.S_IXUSR | ~stat.S_IXGRP | ~stat.S_IXOTH
    os.chmod(fn, existing_permissions | flags)


def get_data_file(name):
    return (resources.files('ros_glint.data') / name).read_text()


def get_ignore_data(name, variables, add_newline=True):
    def get_ignore_data_helper(basename, add_newline=True):
        lines = []
        for s in get_data_file(basename + '.ignore').split('\n'):
            if s == '':
                continue
            if add_newline:
                lines.append(s + '\n')
            else:
                lines.append(s)
        return lines

    ignore_lines = get_ignore_data_helper(name, add_newline)
    for pattern in get_ignore_data_helper(name + '_patterns', add_newline):
        ignore_lines.append(pattern % variables)
    return ignore_lines



from .core import get_linters

from .glinters import package_xml, cmake, python_setup, ros_interfaces, plugins
from .glinters import cmake_installs, cmake_sorting, cmake_pretty
from .glinters import rviz_config, misc

__all__ = ['get_linters', 'package_xml', 'cmake', 'python_setup', 'ros_interfaces', 'plugins',
           'cmake_installs', 'cmake_sorting', 'cmake_pretty',
           'rviz_config', 'misc']



import difflib
from .terminal import color_diff


def get_diff_string(contents0, contents1, filename):
    s = '=' * 5 + str(filename) + '=' * 45 + '\n'
    d = difflib.Differ()
    result = d.compare(contents0.split('\n'), contents1.split('\n'))
    s += '\n'.join(color_diff(result))
    return s



from stylish_cmake_parser import Command, SectionStyle
from ros_introspect.package import MiscPackageFile
from ros_introspect.components.setup_py import create_setup_py, contains_quoted_string, quote_string, unquote_string
from ros_introspect.components.setup_cfg import SetupCFG
from ros_introspect.components.source_code import SourceCode
from ..core import glinter
from ..cmake_ordering import insert_in_order
from ..python_types import get_python_usage, PythonUsage
from .cmake import section_check
from .cmake_installs import install_section_check, InstallType, check_cmake_python_buildtype
from ..util import set_executable
import re

CATKIN_INSTALL_PYTHON_PRENAME = '\n                      '  # newline plus len('catkin_install_python(')
MAIN_IMPORT_PATTERN = re.compile(r'from ([\w\.]+) import.*main.*')
ENTRY_POINT_SCRIPT_TEMPLATE = """#!/usr/bin/env python3

from {pkg}.{lib_name} import main
main()
"""


@glinter
def check_python_marker(package):
    if get_python_usage(package) == PythonUsage.NONE:
        return

    resource_folder = package.root / 'resource'
    marker_path = resource_folder / package.name
    if marker_path.exists():
        return
    resource_folder.mkdir(exist_ok=True)
    with open(marker_path, 'w') as f:
        f.write('')

    package.add_file(MiscPackageFile(marker_path, package))


@glinter
def generate_ament_cmake_python_entry_points(package):
    if package.build_type != 'ament_cmake':
        return
    entry_dict = get_entry_points(package)

    if not entry_dict:
        return
    execs = package.get_source_by_tags('pyscript', 'python')
    for exec in execs:
        ret = exec.search_lines_for_pattern(MAIN_IMPORT_PATTERN)
        if not ret:
            continue
        package_parts = ret[0][0].split('.')
        if package_parts[0] == package.name and package_parts[-1] in entry_dict:
            del entry_dict[package_parts[-1]]

    scripts_dir = package.root / 'scripts'
    scripts_dir.mkdir(exist_ok=True)

    for name, source in entry_dict.items():
        script_fn = scripts_dir / name
        if not script_fn.exists():
            with open(script_fn, 'w') as f:
                f.write(ENTRY_POINT_SCRIPT_TEMPLATE.format(
                    pkg=package.name,
                    lib_name=name,
                ))
            set_executable(script_fn, True)
            package.add_file(SourceCode(script_fn, package))


@glinter
def sync_package_xml_and_setup_py(package):
    if package.setup_py is None and package.build_type != 'ament_python':
        return

    if package.setup_py is None:
        create_setup_py(package)

    for tag in ['version', 'description', 'license']:
        tags = package.manifest.get_elements_by_tags([tag])
        if not tags:
            continue
        value = tags[0].childNodes[0].nodeValue
        package.setup_py.args[tag] = repr(value)

    for maintainer in package.manifest.get_elements_by_tags(['maintainer']):
        # TODO: Expand for author
        # TODO: Joint multiple people?
        package.setup_py.args['maintainer'] = repr(maintainer.childNodes[0].nodeValue)
        package.setup_py.args['maintainer_email'] = repr(maintainer.getAttribute('email'))

    package.setup_py.changed = True


def has_python_library(package_name, py_src):
    for source in py_src:
        parts = source.rel_fn.parts
        if len(parts) > 1 and parts[0] == 'src' and parts[1] == package_name:
            return True
    return False


@glinter
def check_setup_py(package):
    py_src = package.get_source_by_tags(set(), 'python')
    if not py_src and package.build_type not in ['ament_python', 'catkin']:
        return
    if package.setup_py is None:
        if not has_python_library(package.name, py_src) and package.build_type == 'catkin':
            # No library, and no existing setup_py means nothing to write
            return
        create_setup_py(package)

    if package.build_type == 'catkin':
        if 'catkin_python_setup' not in package.cmake.content_map:
            insert_in_order(package.cmake, Command('catkin_python_setup', package.cmake))
    else:
        package.setup_py.include_data_files(['package.xml'])

        # Determine packages/package_dir
        pkg_paths = [src.rel_fn.parent for src in package.get_source_by_tags('init.py')]
        if not pkg_paths:
            return
        package_dir = None
        if all(path.parts[0] == 'src' for path in pkg_paths):
            packages = [path.parts[1] for path in pkg_paths]
            package_dir = 'src'
        else:
            packages = [path.parts[0] for path in pkg_paths]

        if 'packages' not in package.setup_py.args:
            package.setup_py.args['packages'] = []

        for pkg in packages:
            if contains_quoted_string(package.setup_py.args['packages'], pkg):
                continue
            if package.setup_py.declare_package_name and pkg == package.name:
                value = 'package_name'
                if value in package.setup_py.args['packages']:
                    continue
            else:
                value = quote_string(pkg)
            package.setup_py.args['packages'].append(value)
            package.setup_py.changed = True

        if package_dir:
            # Do this after packages just for standard ordering
            if 'package_dir' not in package.setup_py.args:
                package.setup_py.args['package_dir'] = {}

            if not contains_quoted_string(package.setup_py.args['package_dir'].values(), package_dir):
                package.setup_py.args['package_dir'][quote_string('')] = quote_string(package_dir)
                package.setup_py.changed = True


def get_entry_points(package):
    if package.ros_version == 1:
        return {}

    entries = {}
    for source in package.get_source_by_tags({'entry_pt', 'pylib'}):
        entries[source.rel_fn.stem] = source
    return entries


@glinter
def update_python_installs(package):
    # Part 1: Library Setup for ament_cmake
    python_usage = get_python_usage(package)
    if python_usage != PythonUsage.NONE:
        check_cmake_python_buildtype(package)

        if python_usage == PythonUsage.MIXED_CMAKE_PYTHON:
            section_check(package.cmake, ['${PROJECT_NAME}'], 'ament_python_install_package')

    # Part 2: Executables
    execs = package.get_source_by_tags('pyscript', 'python')
    entry_dict = get_entry_points(package)

    if len(execs) == 0 and len(entry_dict) == 0:
        return

    exec_fns = sorted(str(exec.rel_fn) for exec in execs)

    if package.setup_py:
        package_dir = package.setup_py.args.get('package_dir')
        if package_dir:
            package_dir = unquote_string(list(package_dir.values())[0])
    else:
        package_dir = None

    entries = []
    for name, source in entry_dict.items():
        parts = list(source.rel_fn.parent.parts)
        if parts and parts[0] == package_dir:
            parts = parts[1:]
        parts.append(source.rel_fn.stem)
        py_namespace = '.'.join(parts)
        script = f'{name} = {py_namespace}:main'
        entries.append(script)

    if package.build_type == 'catkin':
        cmd = 'catkin_install_python'
        if cmd not in package.cmake.content_map:
            cmake_cmd = Command(cmd, parent=package.cmake)
            cmake_cmd.add_section('PROGRAMS', exec_fns)
            cmake_cmd.add_section('DESTINATION', ['${CATKIN_PACKAGE_BIN_DESTINATION}'],
                                  SectionStyle(CATKIN_INSTALL_PYTHON_PRENAME))
            insert_in_order(package.cmake, cmake_cmd)
        else:
            section_check(package.cmake, exec_fns, cmd, 'PROGRAMS')

    elif package.build_type == 'ament_python':
        if package.setup_cfg is None:
            package.add_file(SetupCFG(package.root / 'setup.cfg', package))

        scripts_folder = '$base/lib/' + package.name + '/scripts'
        package.setup_cfg.ensure('develop', 'script_dir', scripts_folder)
        package.setup_cfg.ensure('install', 'install_scripts', scripts_folder)

        if exec_fns:
            if 'scripts' not in package.setup_py.args:
                package.setup_py.args['scripts'] = []
                package.setup_py.changed = True
            for exec in exec_fns:
                if not contains_quoted_string(package.setup_py.args['scripts'], exec):
                    package.setup_py.args['scripts'].append(quote_string(exec))
                    package.setup_py.changed = True

        if entries:
            if 'entry_points' not in package.setup_py.args:
                package.setup_py.args['entry_points'] = {}

            entry_points = package.setup_py.args['entry_points']

            console_key = contains_quoted_string(entry_points, 'console_scripts')
            if console_key:
                console_scripts = entry_points[console_key]
            else:
                console_scripts = []
                entry_points[quote_string('console_scripts')] = console_scripts

            for entry in entries:
                if not contains_quoted_string(console_scripts, entry):
                    console_scripts.append(quote_string(entry))
                    package.setup_py.changed = True
    elif python_usage == PythonUsage.MIXED_CMAKE_PYTHON:
        install_section_check(package.cmake, exec_fns, InstallType.PYTHON, package.ros_version)



from ros_introspect.components.cmake import ROS_TESTING_FLAGS, is_testing_group
from stylish_cmake_parser import Command, CommandGroup
from enum import IntEnum

BUILD_TARGET_COMMANDS = ['qt5_wrap_cpp', 'add_library', 'add_executable', 'add_rostest',
                         'target_include_directories', 'add_dependencies', 'target_link_libraries',
                         'set_target_properties', 'ament_target_dependencies']
TEST_COMMANDS = [('group', variable) for variable in ROS_TESTING_FLAGS] + \
                ['catkin_download_test_data',
                 'roslint_cpp', 'roslint_python', 'roslint_add_test',
                 'catkin_add_nosetests', 'catkin_add_gtest', 'ament_add_gtest', 'add_rostest_gtest',
                 'ament_lint_auto_find_test_dependencies']
INSTALL_COMMANDS = ['ament_export_targets', 'install', 'catkin_install_python', 'ament_python_install_module',
                    'ament_python_install_package', 'ament_export_include_directories', 'ament_export_libraries',
                    'ament_export_dependencies', 'pluginlib_export_plugin_description_file']

BASE_ORDERING = ['cmake_minimum_required', 'project',
                 ('group', 'CMAKE_C_STANDARD'), ('group', 'CMAKE_CXX_STANDARD'), ('group', 'CMAKE_COMPILER_IS_GNUCXX'),
                 ('set', 'CMAKE'),
                 'set_directory_properties', 'add_compile_options', 'find_package', 'pkg_check_modules',
                 'moveit_build_options',
                 ('set', 'OTHER'),
                 'catkin_generate_virtualenv', 'catkin_python_setup', 'add_definitions',
                 'add_message_files', 'add_service_files', 'add_action_files', 'rosidl_generate_interfaces',
                 'generate_dynamic_reconfigure_options', 'generate_messages', 'catkin_package', 'catkin_metapackage',
                 BUILD_TARGET_COMMANDS + ['include_directories'],
                 ]


class CMakeOrderStyle(IntEnum):
    TEST_FIRST = 1  # Test commands come strictly before install commands
    INSTALL_FIRST = 2  # Test commands come strictly after install commands
    MIXED = 3  # Test and install commands are intermingled
    UNKNOWN = 4  # There are not install commands and test commands


def classify_content_as_test_or_install(content):
    if isinstance(content, CommandGroup) and is_testing_group(content):
        return 'test'
    elif not isinstance(content, Command):
        return
    elif content.command_name in TEST_COMMANDS:
        return 'test'
    elif content.command_name in INSTALL_COMMANDS:
        return 'install'


def get_style(cmake):
    """Examine the contents of the cmake and determine the style."""
    cats = []
    for content in cmake.contents:
        cat = classify_content_as_test_or_install(content)
        if not cat:
            continue
        elif len(cats) == 0 or cats[-1] != cat:
            cats.append(cat)

            if len(cats) > 2:
                return CMakeOrderStyle.MIXED

    if len(cats) == 2:
        if cats[0] == 'install':
            return CMakeOrderStyle.INSTALL_FIRST
        else:
            return CMakeOrderStyle.TEST_FIRST
    return CMakeOrderStyle.UNKNOWN


def get_ordering(style):
    """Given the style, return the correct ordering."""
    if style == CMakeOrderStyle.INSTALL_FIRST:
        return BASE_ORDERING + INSTALL_COMMANDS + TEST_COMMANDS + ['ament_package']
    else:
        return BASE_ORDERING + TEST_COMMANDS + INSTALL_COMMANDS + ['ament_package']


def get_ordering_index(command_name, ordering):
    """
    Given a command name, determine the integer index into the ordering.

    The ordering is a list of strings and arrays of strings.

    If the command name matches one of the strings in the inner arrays,
    the index of the inner array is returned.

    If the command name matches one of the other strings, its index is returned.

     Otherwise, the length of the ordering is returned (putting non-matches at the end)
    """
    for i, o in enumerate(ordering):
        if isinstance(o, list):
            if command_name in o:
                return i
        elif command_name == o:
            return i
    if command_name:
        # TODO: Raise Warning
        print(f'\tUnsure of ordering for {command_name}')
    return len(ordering)


def get_sort_key(content, anchors, ordering):
    """
    Given a piece of cmake content, return a tuple representing its sort_key.

    The first element of the tuple is the ordering_index of the content.
    The second element is an additional variable used for sorting among elements with the same ordering_index

    Most notably, we want all build commands with a particular library/executable to be grouped together.
    In that case, we use the anchors parameter, which is an ordered list of all the library/executables in the file.
    Then, the second variable is a tuple itself, with the first element being the index of library/executable in the
    anchors list, and the second is an integer representing the canonical order of the build commands.
    """
    if content is None:
        return len(ordering) + 1, None
    index = None
    key = ()
    if content.__class__ == CommandGroup:
        key_token = ()
        for token in content.initial_cmd.get_tokens(include_name=True):
            if token == 'NOT':
                continue
            key_token = token
            break
        index = get_ordering_index(('group', key_token), ordering)
    elif content.command_name == 'set':
        token = content.get_tokens(include_name=True)[0]
        if token.startswith('CMAKE_'):
            set_type = 'CMAKE'
        else:
            set_type = 'OTHER'
        index = get_ordering_index(('set', set_type), ordering)
    else:  # Command
        index = get_ordering_index(content.command_name, ordering)
        if content.command_name in BUILD_TARGET_COMMANDS:
            token = content.first_token()
            key = anchors.index(token), BUILD_TARGET_COMMANDS.index(content.command_name)
        elif content.command_name == 'include_directories' and 'include_directories' in anchors:
            key = -1, anchors.index('include_directories')
        elif content.command_name == 'find_package':
            token = content.get_tokens()[0]
            if token == 'catkin' or token.startswith('ament_cmake'):
                key = -1, token  # Force ament_cmake/catkin to come first
            else:
                key = 0, token
    return index, key


def get_ordered_build_targets(cmake):
    targets = []
    for content in cmake.contents:
        if content.__class__ != Command:
            continue
        if content.command_name == 'include_directories':
            targets.append('include_directories')
            continue
        elif content.command_name not in BUILD_TARGET_COMMANDS:
            continue
        token = content.first_token()
        if token not in targets:
            targets.append(token)
    return targets


def get_insertion_index(cmake, cmd):
    anchors = get_ordered_build_targets(cmake)
    ordering = get_ordering(CMakeOrderStyle.TEST_FIRST)

    new_key = get_sort_key(cmd, anchors, ordering)
    i_index = 0

    for i, content in enumerate(cmake.contents):
        if isinstance(content, str):
            continue
        key = get_sort_key(content, anchors, ordering)
        if key <= new_key:
            i_index = i + 1
        elif key[0] != len(ordering):
            return i_index
    return len(cmake.contents)


def insert_in_order(cmake, cmd):
    cmake.insert(cmd, get_insertion_index(cmake, cmd))



import re

from ..core import glinter
from ..cmake_ordering import insert_in_order
from ..util import get_ignore_data

from ros_introspect.package import DependencyType
from ros_introspect.components.source_code import CPP_INCLUDE1, CPP_INCLUDE2
from ros_introspect.ros_resources import ROSResources
from stylish_cmake_parser import Command, CommandGroup, SectionStyle
ALL_SPACES = re.compile(r' +')
NEWLINE_PLUS_2 = '\n  '
NEWLINE_PLUS_4 = '\n    '
NEWLINE_PLUS_8 = '\n        '


def remove_empty_strings(a):
    return list(filter(lambda x: x != '', a))


def remove_cmake_command_comments_helper(command, ignorables, replacement=''):
    for i, section in enumerate(command.sections):
        if not isinstance(section, str):
            continue
        for ignorable in ignorables:
            while ignorable in command.sections[i]:
                command.mark_changed()
                command.sections[i] = command.sections[i].replace(ignorable, replacement)
    if command.changed:
        command.sections = remove_empty_strings(command.sections)
        if command.sections == ['\n']:
            command.sections = []


def remove_cmake_comments_helper(cmake, ignorables, replacement=''):
    # Remove Raw Strings
    for ignorable in ignorables:
        while ignorable in cmake.contents:
            i = cmake.contents.index(ignorable)
            if i > 1:
                one_before = cmake.contents[i - 1]
                two_before = cmake.contents[i - 2]
                if isinstance(one_before, str) and isinstance(two_before, str):
                    if ALL_SPACES.match(one_before) and two_before[-1] == '\n':
                        cmake.contents = cmake.contents[:i - 1] + cmake.contents[i + 1:]
                        cmake.mark_changed()
                        continue

            cmake.contents = cmake.contents[:i] + cmake.contents[i + 1:]
            cmake.mark_changed()

    # Remove from Commands/CommandGroups
    for content in cmake.contents:
        if isinstance(content, Command):
            remove_cmake_command_comments_helper(content, ignorables, replacement)
        elif isinstance(content, CommandGroup):
            remove_cmake_comments_helper(content.contents, ignorables, replacement)
    cmake.contents = remove_empty_strings(cmake.contents)


@glinter
def remove_empty_cmake_lines(package):
    cmake = package.cmake
    for i, content in enumerate(cmake.contents[:-2]):
        if str(content)[-1] == '\n' and cmake.contents[i + 1] == '\n' and cmake.contents[i + 2] == '\n':
            cmake.contents[i + 1] = ''
            cmake.mark_changed()
    cmake.contents = remove_empty_strings(cmake.contents)


@glinter
def remove_boilerplate_cmake_comments(package):
    if not package.cmake:
        return

    ignorables = get_ignore_data('cmake', {'package': package.name})
    remove_cmake_comments_helper(package.cmake, ignorables)
    remove_empty_cmake_lines(package)


def get_command_section(cmake, cmd_name, section_name):
    """Return the first matching section from the properly named command"""
    for cmd in cmake.content_map[cmd_name]:
        s = cmd.get_section(section_name)
        if s:
            return s


def ensure_section_values(cmake, section, items, alpha_order=True, ignore_quoting=False):
    """Ensure the CMake command with the given section has all of the values in items."""
    existing = cmake.resolve_variables(section.values)

    needed_items = []
    for item in items:
        if item in existing or item in section.values:
            continue

        if ignore_quoting:
            quoted = f'"{item}"'
            if quoted in existing or quoted in section.values:
                continue

        # TODO: Should maybe follow quote style
        needed_items.append(item)

    if needed_items:
        section.add_values(needed_items, alpha_order)
        return True
    return False


def section_check(cmake, items, cmd_name, section_name='', zero_okay=False, alpha_order=True, ignore_quoting=False):
    """Ensure there's a CMake command of the given type with the given section name and items."""

    if len(items) == 0 and not zero_okay:
        return

    section = get_command_section(cmake, cmd_name, section_name)

    if not section:
        if cmake.content_map[cmd_name]:
            cmd = cmake.content_map[cmd_name][0]
        else:
            cmd = Command(cmd_name, parent=cmake)
            insert_in_order(cmake, cmd)

    if section is None:
        cmd.add_section(section_name, sorted(items))
    else:
        ensure_section_values(cmake, section, items, alpha_order, ignore_quoting)
    return


def targeted_section_check(cmake, cmd_name, section_name, items, style=None):
    def get_targeted_command(target_name):
        for cmd in cmake.content_map[cmd_name]:
            tokens = cmd.get_tokens()
            if tokens and target_name == tokens[0]:
                return cmd

    for target in cmake.get_libraries() + cmake.get_executables():
        cmd = get_targeted_command(target)
        if cmd is None:
            cmd = Command(cmd_name, parent=cmake)
            if section_name:
                cmd.add_section('', [target])
                cmd.add_section(section_name, items, style)
            else:
                cmd.add_section(section_name, [target] + items, style)
            insert_in_order(cmake, cmd)
        else:
            section = cmd.get_section(section_name)
            if not section:
                continue
            if style:
                section.style = style
            ensure_section_values(cmake, section, items, alpha_order=False)


def get_multiword_section(cmd, words):
    """Find a section that matches the last word, assuming all the previous sections matched the other words.

    Our definition of a CMake command section is ONE all-caps word followed by tokens.
    Installing stuff requires these weird TWO word sections (i.e. ARCHIVE DESTINATION).

    Ergo, we need to find the section that matches the second word, presuming the section
    before matched the first word.
    """
    i = 0
    for section in cmd.get_real_sections():
        if section.name == words[i]:
            # One word matches
            if i < len(words) - 1:
                # If there are more words, increment the counter
                i += 1
            else:
                # Otherwise, we matched all the words. Return this section
                return section
        else:
            # If the word doesn't match, we need to start searching from the first word again
            i = 0


def check_complex_section(cmd, key, value):
    """Find the section matching the key and ensure the value is in it.

    Key could be multiple words, see get_multiword_section.
    If the appropriate section is not found, it adds it.
    """
    words = key.split()
    if len(words) == 1:
        section = cmd.get_section(key)
    else:
        section = get_multiword_section(cmd, words)

    if section:
        if value not in section.values:
            section.add(value)
    else:
        cmd.add_section(key, [value], SectionStyle(NEWLINE_PLUS_8))


def install_cmake_dependencies(package, dependencies, check_catkin_pkg=True):
    if not dependencies:
        return

    cmake = package.cmake
    if package.ros_version == 1:
        if len(cmake.content_map['find_package']) == 0:
            cmd = Command('find_package', parent=cmake)
            # TODO: Standard indentation
            cmd.add_section('', ['catkin'])
            cmd.add_section('REQUIRED')
            cmake.add_command(cmd)

        for cmd in cmake.content_map['find_package']:
            tokens = cmd.get_tokens()
            if not tokens or tokens[0] != 'catkin':
                continue

            req_sec = cmd.get_section('REQUIRED')
            if not req_sec:
                continue
            section = cmd.get_section('COMPONENTS')
            if section is None and req_sec.values:
                section = req_sec  # Allow packages to be listed without COMPONENTS keyword
            if section is None:
                cmd.add_section('COMPONENTS', sorted(dependencies))
            else:
                existing = cmake.resolve_variables(section.values)
                needed_items = dependencies - set(existing)
                if needed_items:
                    section.add_values(needed_items)
        if check_catkin_pkg:
            section_check(cmake, dependencies, 'catkin_package', 'CATKIN_DEPENDS')
    else:
        existing = set()
        for cmd in cmake.content_map['find_package']:
            existing.update(cmd.get_tokens())

        for pkg in sorted(dependencies - existing):
            cmd = Command('find_package', parent=cmake)
            cmd.add_section('', [pkg])
            cmd.add_section('REQUIRED')
            insert_in_order(cmake, cmd)


@glinter
def check_cmake_dependencies(package):
    if not package.cmake:
        return
    dependencies = package.get_dependencies(DependencyType.BUILD)
    install_cmake_dependencies(package, dependencies)


@glinter
def check_generators(package):
    all_interfaces = package.get_ros_interfaces()
    if not package.cmake or not all_interfaces:
        return

    cmake = package.cmake
    if package.ros_version == 1:
        section_check(cmake, [m.fn for m in package.messages], 'add_message_files', 'FILES')
        section_check(cmake, [s.fn for s in package.services], 'add_service_files', 'FILES')
        section_check(cmake, [a.fn for a in package.actions], 'add_action_files', 'FILES')
        section_check(cmake, ['message_generation'], 'find_package', 'COMPONENTS')
        section_check(cmake, ['message_runtime'], 'catkin_package', 'CATKIN_DEPENDS')

        generate_cmd = 'generate_messages'
    else:
        generate_cmd = 'rosidl_generate_interfaces'

        if not cmake.content_map[generate_cmd]:
            cmd = Command('rosidl_generate_interfaces', parent=cmake)
            cmd.add_section('${PROJECT_NAME}')
            insert_in_order(cmake, cmd)

        rel_fns = [str(ros_interface.rel_fn) for ros_interface in all_interfaces]
        section_check(cmake, rel_fns, generate_cmd, '', ignore_quoting=True)

        install_cmake_dependencies(package, {'rosidl_default_generators'})
        section_check(cmake, ['rosidl_default_runtime'], 'ament_export_dependencies', '')

    msg_deps = set()
    for ros_interface in all_interfaces:
        msg_deps |= ros_interface.get_dependencies(DependencyType.BUILD)

    if msg_deps:
        section_check(cmake, msg_deps, generate_cmd, 'DEPENDENCIES', zero_okay=True)
    else:
        section_check(cmake, msg_deps, generate_cmd, zero_okay=True)


@glinter
def check_includes(package):
    if not package.cmake:
        return

    has_cpp = any(source_file.language == 'c++' for source_file in package.source_code)
    has_header_files = any(package.get_source_by_tags('header'))

    if not has_cpp:
        return

    cmake = package.cmake

    if package.ros_version == 1:
        if has_header_files:
            section_check(cmake, ['include'], 'catkin_package', 'INCLUDE_DIRS')
            section_check(cmake, ['include'], 'include_directories', alpha_order=False)

        section_check(cmake, ['${catkin_INCLUDE_DIRS}'], 'include_directories', alpha_order=False)
    else:
        include_directories = []
        if has_header_files:
            include_directories.append('$<BUILD_INTERFACE:${CMAKE_CURRENT_SOURCE_DIR}/include>')

        include_directories.append('$<INSTALL_INTERFACE:include>')

        targeted_section_check(package.cmake, 'target_include_directories', 'PUBLIC', include_directories,
                               SectionStyle(name_val_sep='\n  ', val_sep='\n  '))


@glinter
def target_catkin_libraries(package):
    if not package.cmake:
        return

    if package.ros_version == 1:
        deps = ['${catkin_LIBRARIES}']
        command = 'target_link_libraries'
        style = None
    else:
        deps = sorted(package.get_dependencies(DependencyType.BUILD))
        command = 'ament_target_dependencies'
        style = SectionStyle(prename='', val_sep='\n  ')
    targeted_section_check(package.cmake, command, '', deps, style)


@glinter
def check_library_setup(package):
    if not package.cmake:
        return
    if package.ros_version != 1:
        return

    section_check(package.cmake, package.cmake.get_libraries(),
                  'catkin_package', 'LIBRARIES')


def get_target_build_rules(cmake):
    targets = {}
    targets.update(cmake.get_build_rules('add_library'))
    targets.update(cmake.get_build_rules('add_executable'))
    return targets


def get_msg_dependencies_from_source(package, sources):
    resources = ROSResources.get()
    deps = set()
    for rel_fn in sources:
        srcs = [sc for sc in package.source_code if str(sc.rel_fn) == rel_fn]
        if not srcs:
            continue
        src = srcs[0]
        for pkg, name in src.search_lines_for_pattern(CPP_INCLUDE1):
            if len(name) == 0 or name[-2:] != '.h':
                continue
            name = name.replace('.h', '')
            if resources.is_message(pkg, name) or resources.is_service(pkg, name):
                deps.add(pkg)
            elif pkg == package.name and any(name == ros_int.name for ros_int in package.get_ros_interfaces()):
                deps.add(pkg)
        for pkg, gen_type, name in src.search_lines_for_pattern(CPP_INCLUDE2):
            if gen_type == 'msg' or gen_type == 'srv':
                deps.add(pkg)
    if package.dynamic_reconfig:
        deps.add(package.name)
    return sorted(deps)


def get_matching_add_depends(cmake, search_target):
    valid_targets = {search_target}
    alt_target = cmake.resolve_variables(search_target)
    if alt_target != search_target:
        valid_targets.add(alt_target)

    for cmd in cmake.content_map['add_dependencies']:
        target = cmd.first_token()
        if target in valid_targets:
            return cmd
        resolved_target = cmake.resolve_variables(target)
        if resolved_target in valid_targets:
            return cmd


@glinter
def check_exported_dependencies(package):
    if not package.cmake or package.ros_version == 2:
        return
    cmake = package.cmake
    targets = get_target_build_rules(cmake)

    for target, sources in targets.items():
        deps = get_msg_dependencies_from_source(package, sources)
        if len(deps) == 0:
            continue

        if package.name in deps:
            self_depend = True
            if len(deps) == 1:
                cat_depend = False
            else:
                cat_depend = True
        else:
            self_depend = False
            cat_depend = True

        add_deps = get_matching_add_depends(package.cmake, target)
        add_add_deps = False

        if add_deps is None:
            add_deps = Command('add_dependencies', parent=cmake)
            add_add_deps = True  # Need to wait to add the command for proper sorting

        if len(add_deps.sections) == 0:
            add_deps.add_section('', [target])

        section = add_deps.sections[0]
        if cat_depend and '${catkin_EXPORTED_TARGETS}' not in section.values:
            section.add('${catkin_EXPORTED_TARGETS}')
        if self_depend:
            tokens = [cmake.resolve_variables(s) for s in section.values]
            key = '${%s_EXPORTED_TARGETS}' % package.name
            if key not in tokens:
                section.add(key)

        if add_add_deps:
            insert_in_order(cmake, add_deps)


@glinter
def remove_old_style_cpp_dependencies(package):
    def remove_pattern(section, pattern):
        prev_len = len(section.values)
        section.values = [v for v in section.values if pattern not in v]
        return prev_len != len(section.values)

    if not package.cmake:
        return
    global_changed = False
    targets = get_target_build_rules(package.cmake)
    for target in targets:
        add_deps = get_matching_add_depends(package.cmake, target)
        if add_deps is None or len(add_deps.sections) == 0:
            continue

        section = add_deps.sections[0]
        changed = remove_pattern(section, '_generate_messages_cpp')
        changed = remove_pattern(section, '_gencpp') or changed
        changed = remove_pattern(section, '_gencfg') or changed
        if changed:
            add_deps.changed = True
            global_changed = True
    if global_changed:
        check_exported_dependencies(package)



from ..core import glinter
from ..config import get_config
from ..cmake_ordering import CMakeOrderStyle, get_ordered_build_targets, get_ordering, get_sort_key, get_style


def get_clusters(cmake, desired_style):
    """Return a list of clusters where each cluster is an array of strings with a Command/CommandGroup at the end.

    The clusters are sorted according to the desired style.
    The strings are grouped at the beginning to maintain the newlines and indenting before each Command.

    Can also return None if everything is already sorted
    """
    anchors = get_ordered_build_targets(cmake)
    ordering = get_ordering(desired_style)
    clusters = []
    current = []
    prev_key = None
    needs_sorting = False

    for content in cmake.contents:
        current.append(content)
        if isinstance(content, str):
            continue
        key = get_sort_key(content, anchors, ordering)
        if prev_key and key < prev_key:
            needs_sorting = True
        prev_key = key
        clusters.append((key, current))
        current = []
    if len(current) > 0:
        last_key = get_sort_key(None, anchors, ordering)
        if prev_key and last_key < prev_key:
            needs_sorting = True
        clusters.append((last_key, current))

    if not needs_sorting:
        return
    return [kv[1] for kv in sorted(clusters, key=lambda kv: kv[0])]


def enforce_ordering(cmake, default_style=None):
    existing_style = get_style(cmake)
    if default_style in [CMakeOrderStyle.INSTALL_FIRST, CMakeOrderStyle.TEST_FIRST]:
        desired_style = default_style
    elif existing_style in [CMakeOrderStyle.INSTALL_FIRST, CMakeOrderStyle.TEST_FIRST]:
        desired_style = existing_style
    else:
        desired_style = CMakeOrderStyle.TEST_FIRST

    # Recurse through CommandGroups
    for group in cmake.content_map['group']:
        enforce_ordering(group.contents, default_style)

    clusters = get_clusters(cmake, desired_style)
    if not clusters:  # Already sorted
        return

    cmake.contents = []
    for contents in clusters:
        cmake.contents += contents

    cmake.mark_changed()


@glinter
def enforce_cmake_ordering(package, config=None):
    if not package.cmake:
        return
    if config is None:
        config = get_config()
    default_style_s = config.get('cmake_style', 'UNKNOWN')
    default_style = CMakeOrderStyle[default_style_s.upper()]

    enforce_ordering(package.cmake, default_style)



from ..core import glinter
from ros_introspect.components.ros_interface import PRIMITIVES

STANDARD = {
    'Header': 'std_msgs'
}


@glinter
def clean_whitespace_from_interface_definition(package):
    # Formerly remove_trailing_whitespace_from_generators
    for interface in package.get_ros_interfaces():
        interface.force_regeneration()


@glinter
def fill_in_msg_package_names(package):
    interfaces = package.get_ros_interfaces()
    all_names = {ros_i.name for ros_i in interfaces}

    for interface in interfaces:
        for section in interface.sections:
            for field in section.fields:
                if '/' in field.type or field.type in PRIMITIVES:
                    continue

                if field.type in STANDARD:
                    field.type = STANDARD[field.type] + '/' + field.type
                    interface.changed = True
                elif field.type in all_names:
                    field.type = package.name + '/' + field.type
                    interface.changed = True



from stylish_cmake_parser import Command, CommandGroup
from ..core import glinter
from .cmake import NEWLINE_PLUS_2, NEWLINE_PLUS_4, NEWLINE_PLUS_8
from .python_setup import CATKIN_INSTALL_PYTHON_PRENAME

SHOULD_ALPHABETIZE = ['COMPONENTS', 'DEPENDENCIES', 'FILES', 'CATKIN_DEPENDS']


def set_style_attribute(section, attribute, new_value):
    existing_value = getattr(section.style, attribute)
    if existing_value == new_value:
        return
    setattr(section.style, attribute, new_value)
    section.mark_changed()


@glinter
def alphabetize_cmake_sections(package):
    if not package.cmake:
        return

    def alphabetize_sections_helper(cmake):
        for content in cmake.contents:
            if isinstance(content, Command):
                for section in content.get_real_sections():
                    if section.name in SHOULD_ALPHABETIZE:
                        sorted_values = sorted(section.values)
                        if sorted_values != section.values:
                            section.values = sorted_values
                            content.mark_changed()
            elif isinstance(content, CommandGroup):
                alphabetize_sections_helper(content.contents)

    alphabetize_sections_helper(package.cmake)


@glinter
def prettify_catkin_package_cmd(package):
    if not package.cmake:
        return
    for cmd in package.cmake.content_map['catkin_package']:
        for section in cmd.get_real_sections():
            set_style_attribute(section, 'prename', NEWLINE_PLUS_4)


@glinter
def prettify_package_lists(package):
    if not package.cmake:
        return
    acceptable_styles = [(NEWLINE_PLUS_8, NEWLINE_PLUS_8), (NEWLINE_PLUS_4, NEWLINE_PLUS_8)]

    for cmd_name, section_name in [('find_package', 'COMPONENTS'),
                                   ('catkin_package', 'CATKIN_DEPENDS')]:
        for cmd in package.cmake.content_map[cmd_name]:
            for section in cmd.get_real_sections():
                if section.name != section_name:
                    continue
                n = len(str(section))
                if n > 120:
                    key = section.style.name_val_sep, section.style.val_sep
                    if key not in acceptable_styles:
                        section.style.name_val_sep = NEWLINE_PLUS_8
                        section.style.val_sep = NEWLINE_PLUS_8
                        section.mark_changed()


@glinter
def prettify_msgs_srvs(package):
    if not package.cmake:
        return
    # ROS 1 version
    for cmd in package.cmake.content_map['add_message_files'] + package.cmake.content_map['add_service_files']:
        for section in cmd.get_real_sections():
            if len(section.values) > 1:
                set_style_attribute(section, 'name_val_sep', NEWLINE_PLUS_4)
                set_style_attribute(section, 'val_sep', NEWLINE_PLUS_4)
    # ROS 2 version
    for cmd in package.cmake.content_map['rosidl_generate_interfaces']:
        for section in cmd.get_real_sections():
            if section.name == '' and section == cmd.sections[0]:
                set_style_attribute(section, 'val_sep', NEWLINE_PLUS_4)
            elif section.name == 'DEPENDENCIES':
                set_style_attribute(section, 'prename', NEWLINE_PLUS_2)
                set_style_attribute(section, 'val_sep', NEWLINE_PLUS_4)
                set_style_attribute(section, 'name_val_sep', NEWLINE_PLUS_4)


@glinter
def prettify_installs(package):
    if not package.cmake:
        return
    for cmd in package.cmake.content_map['install']:
        cmd.sections = [s for s in cmd.sections if not isinstance(s, str)]
        if not cmd.sections:
            continue

        if '\n' in cmd.sections[0].style.prename:
            new_style = cmd.sections[0].style.prename
        else:
            new_style = NEWLINE_PLUS_8

        zeroed = False
        for section in cmd.sections[1:]:
            if len(section.values) == 0:
                set_style_attribute(section, 'prename', new_style)
                zeroed = True
            elif not zeroed:
                set_style_attribute(section, 'prename', new_style)
            else:
                set_style_attribute(section, 'prename', '')

    for cmd in package.cmake.content_map['catkin_install_python']:
        section = cmd.sections[1]
        set_style_attribute(section, 'prename', CATKIN_INSTALL_PYTHON_PRENAME)


@glinter
def prettify_command_groups(package):
    if not package.cmake:
        return

    def pretty_groups_helper(cmake, indent=0):
        for group in cmake.content_map['group']:
            pretty_groups_helper(group.contents, indent + 2)

        if indent == 0:
            return
        for i, content in enumerate(cmake.contents):
            if isinstance(content, Command) and i > 0 and isinstance(cmake.contents[i-1], str):
                prev = cmake.contents[i-1]
                while '\n ' in prev:
                    prev = prev.replace('\n ', '\n')
                revised = prev.replace('\n', '\n' + (' ' * indent))
                if prev != revised:
                    cmake.contents[i - 1] = revised
                    cmake.mark_changed()

    pretty_groups_helper(package.cmake)



import collections


_glinter_functions = collections.OrderedDict()


# Decorator function for gathering all the functions
def glinter(f):
    _glinter_functions[f.__name__] = f
    return f


def get_linters():
    return _glinter_functions



from enum import IntEnum


class PythonUsage(IntEnum):
    NONE = 0
    PURE_PYTHON = 1
    MIXED_CMAKE_PYTHON = 2


def get_python_usage(package):
    # If there is no Python library, return NONE
    if not package.get_source_by_tags('pylib') and package.build_type != 'ament_python':
        return PythonUsage.NONE

    # If build type is ament_cmake or ament_cmake_python
    if 'ament_cmake' in package.build_type:
        return PythonUsage.MIXED_CMAKE_PYTHON

    return PythonUsage.PURE_PYTHON



import collections
import re

from ros_introspect.components.plugin_xml import PluginXML

from ..core import glinter
from .cmake import install_cmake_dependencies, section_check

PLUGIN_PATTERN = r'PLUGINLIB_EXPORT_CLASS\(([^:]+)::([^,]+),\s*([^:]+)::([^,]+)\)'
PLUGIN_RE = re.compile(PLUGIN_PATTERN)


def get_plugin_xmls(package_xml):
    """Return a mapping from the package name to a list of the relative path(s) for the plugin xml(s)."""
    xmls = collections.defaultdict(list)
    export = package_xml.root.getElementsByTagName('export')
    for export_tag in export:
        for node in export_tag.childNodes:
            if node.nodeType == package_xml.root.ELEMENT_NODE:
                plugin = node.getAttribute('plugin').replace('${prefix}/', '')
                xmls[node.nodeName].append(plugin)
    return xmls


def add_to_package_xml(package_xml, plugin_xml, package_name):
    export_tags = package_xml.root.getElementsByTagName('export')
    if not export_tags:
        # Create export tag
        export_tag = package_xml.create_new_tag('export')
    else:
        export_tag = export_tags[0]

    attr = '${prefix}/' + str(plugin_xml.rel_fn)
    for tag in export_tag.childNodes:
        if tag.nodeName != package_name:
            continue
        plugin = tag.attributes.get('plugin')
        if plugin and plugin.value == attr:
            # No change needed
            return

    # Plugin not found
    pe = package_xml.tree.createElement(package_name)
    pe.setAttribute('plugin', attr)

    if not export_tag.childNodes:
        export_tag.appendChild(package_xml.tree.createTextNode('\n    '))
        export_tag.appendChild(pe)
        export_tag.appendChild(package_xml.tree.createTextNode('\n  '))
    else:
        all_elements = []
        all_elements.append(package_xml.create_new_tab_element(2))
        all_elements.append(pe)
        export_tag.childNodes = export_tag.childNodes[:-1] + all_elements + export_tag.childNodes[-1:]
    package_xml.changed = True


@glinter
def check_plugins(package):
    """Check that all the plugins are properly defined.

    We have three dictionaries
      * The plugins that are defined by macros (defined_macros)
      * The plugins that have associated configuration files (existing_plugins)
      * The plugins that are linked by the manifest. (plugin_xml_by_package)
    First, we reconcile the macros with the files.
    Then we handle the manifest.
    Then we make sure that the specific classes are in the configurations
    """
    if not package.cmake:
        return

    existing_plugins = collections.defaultdict(list)
    for xml in package.plugin_xml:
        for parent_pkg in xml.parent_pkgs:
            existing_plugins[parent_pkg].append(xml)

    defined_plugins = get_plugin_xmls(package.package_xml)
    build_rules = package.cmake.get_build_rules('add_library', resolve_target_name=True)

    def contains_library(xmls, library, pkg, name):
        for xml in xmls:
            if xml.contains_library(library, pkg, name):
                return True
        return False

    def lookup_library(rel_fn):
        for library, deps in build_rules.items():
            if rel_fn in deps:
                return library

    for source_file in package.source_code:
        plugin_info = source_file.search_lines_for_pattern(PLUGIN_RE)
        if not plugin_info:
            continue
        library = lookup_library(str(source_file.rel_fn))
        if library is None:
            continue
        for pkg1, name1, pkg2, name2 in plugin_info:
            # Create file if needed
            if pkg2 not in existing_plugins:
                xml_filename = f'{pkg2}_plugins.xml'
                full_path = package.root / xml_filename
                p_xml = PluginXML(full_path, package)
                package.plugin_xml.append(p_xml)
                existing_plugins[pkg2] = [p_xml]

            # Make sure plugins are properly exported
            for plugin_xml in existing_plugins[pkg2]:
                if package.ros_version == 1:
                    if plugin_xml.rel_fn not in defined_plugins[pkg2]:
                        add_to_package_xml(package.package_xml, plugin_xml, pkg2)
                else:
                    section_check(package.cmake, [pkg2, str(plugin_xml.rel_fn)],
                                  'pluginlib_export_plugin_description_file',
                                  '')

            # Make sure the class is in the files
            if not contains_library(existing_plugins[pkg2], library, pkg1, name1):
                # insert into first
                xml = existing_plugins[pkg2][0]
                xml.insert_new_class(library, pkg1, name1, pkg2, name2)

        if package.ros_version > 1:
            install_cmake_dependencies(package, {'ament_cmake_ros'})



import inspect
import pooch
import pytest
from ros_glint import get_linters
from ros_glint.diff import get_diff_string
from ros_glint.terminal import color_text
from zip_interface import get_test_cases
from betsy_ros import ROSInterface

from ros_introspect import Package, ROSResources

URL_TEMPLATE = 'https://github.com/DLu/roscompile_test_data/raw/{}/test_data.zip'
TEST_DATA = [
    # (branch, known_hash)
    ('ros1', '84bf97b92d2ceb8d4369ed0d6fd8a01fabaf3006034fb7710d448b8cfe9885d9'),
    ('ros2', 'cf23d832e0cf56c8e160c0178af81e865c615fb330fb26604ddf031fc51a280a'),
]

linters = get_linters()


def run_case(test_config, cases):
    resources = ROSResources.get()

    with cases[test_config['in']] as pkg_in:
        if test_config['in'] == test_config['out']:
            pkg_out = pkg_in.copy()
        else:
            pkg_out = cases[test_config['out']]

        root = pkg_in.root
        pkg_obj = Package(root)
        local_config = test_config.get('config', {})

        # Initialize ROS Resources
        resources.packages = set(test_config.get('pkgs', []))
        resources.messages = set()
        for msg in test_config.get('msgs', []):
            parts = msg.split('/')
            resources.messages.add(ROSInterface(parts[0], 'msg', parts[1]))

        # Run Functions
        for function_name in test_config['functions']:
            if function_name not in linters:
                pytest.skip(f'Missing linter: {function_name}')

            fne = linters[function_name]
            if 'config' in inspect.getfullargspec(fne).args:
                fne(pkg_obj, config=local_config)
            else:
                fne(pkg_obj)

        components_with_changes = [comp.rel_fn for comp in pkg_obj if comp.changed]
        pkg_obj.save()

        s = '{:25} >> {:25} {}'.format(test_config['in'], test_config['out'],
                                       ','.join(test_config['functions']))
        print(color_text(s, 'BLUE', bright=True))

        # Compute the differences
        filenames_in = pkg_in.get_filenames()
        filenames_out = pkg_out.get_filenames()
        problems = []

        for filename in sorted(filenames_in - filenames_out):
            print(get_diff_string(pkg_in.get_contents(filename).rstrip(), '', filename))
            problems.append(f"File {filename} should have been deleted but wasn't")
        for filename in sorted(filenames_out - filenames_in):
            print(get_diff_string('', pkg_out.get_contents(filename).rstrip(), filename))
            problems.append(f"File {filename} should have been generated but wasn't")
        for filename in sorted(filenames_in.intersection(filenames_out)):
            contents_in = pkg_in.get_contents(filename).rstrip()
            contents_out = pkg_out.get_contents(filename).rstrip()

            if contents_in == contents_out:
                if filename in components_with_changes and test_config['in'] == test_config['out']:
                    problems.append(f'File {filename} has invisible changes')
            else:
                print(get_diff_string(contents_in, contents_out, filename))
                problems.append('The contents of {} do not match!'.format(filename))

        assert not problems, ', '.join(problems)


parameters = []
test_ids = []

for branch, known_hash in TEST_DATA:
    file_path = pooch.retrieve(URL_TEMPLATE.format(branch), known_hash=known_hash)
    config, test_data = get_test_cases(file_path)

    for i, test_config in enumerate(config):
        parameters.append((test_config, test_data))
        test_ids.append(f'{branch}_test_{i:03d}')


@pytest.mark.parametrize('test_config, test_data', parameters, ids=test_ids)
def test_from_zip(test_config, test_data):
    run_case(test_config, test_data)


@pytest.mark.parametrize('test_config, test_data', parameters, ids=test_ids)
def test_idempotency_from_zip(test_config, test_data):
    test_config_x = dict(test_config)
    test_config_x['in'] = test_config_x['out']
    run_case(test_config_x, test_data)



import pathlib
import yaml

DOT_ROS_FOLDER = pathlib.Path('~/.ros').expanduser()
POSSIBLE_CONFIG_FILES = ['roscompile.yaml', 'glint.yaml']
CONFIG = None


def get_config():
    global CONFIG
    if CONFIG is None:
        for possible_config_file in POSSIBLE_CONFIG_FILES:
            path = DOT_ROS_FOLDER / possible_config_file
            if path.exists():
                CONFIG = yaml.safe_load(open(path))
                return CONFIG
        CONFIG = {}
    return CONFIG



import collections
import os
import pathlib
import tempfile
import yaml
import zipfile

from ros_introspect.finder import walk
from ros_glint.util import set_executable


class ROSCompilePackageFiles:
    def __init__(self, package_name, pkg_files, executables):
        self.package_name = package_name
        self.is_written = False
        self.tempdir = None
        self.root = None
        self.pkg_files = pkg_files
        self.executables = executables

    def copy(self):
        return ROSCompilePackageFiles(self.package_name, self.pkg_files, self.executables)

    def __enter__(self):
        self.tempdir = tempfile.TemporaryDirectory()
        self.root = pathlib.Path(self.tempdir.name)
        self.write()
        self.is_written = True
        return self

    def __exit__(self, a_type, value, traceback):
        self.is_written = False
        self.root = None
        self.tempdir = None

    def get_filenames(self):
        if self.is_written:
            the_files = []
            for subpath in walk(self.root):
                the_files.append(subpath)
            return set(the_files)
        else:
            return set(self.pkg_files.keys())

    def get_contents(self, filename):
        if self.is_written:
            full_path = self.root / filename
            if full_path.exists():
                return open(full_path).read().replace('\r\n', '\n')
        elif filename in self.pkg_files:
            return self.pkg_files[filename].replace('\r\n', '\n')

    def write(self):
        for fn, contents in self.pkg_files.items():
            outfile = self.root / fn
            outfile.parent.mkdir(exist_ok=True, parents=True)
            with open(outfile, 'w') as f:
                f.write(contents)
            if fn in self.executables:
                set_executable(outfile, True)

    def __repr__(self):
        return self.package_name


def get_test_cases(zip_filename):
    file_data = collections.defaultdict(dict)
    zf = zipfile.ZipFile(zip_filename)
    config = None
    executables = set()
    for file in zf.filelist:
        if file.filename[-1] == '/':
            continue
        if file.filename == 'list_o_tests.yaml':
            config = yaml.safe_load(zf.read(file))
            continue
        parts = file.filename.split(os.path.sep)
        package = parts[0]
        path = pathlib.Path(os.path.join(*parts[1:]))
        file_data[package][path] = zf.read(file).decode()
        if (file.external_attr >> 16) & 0o111:
            executables.add(path)

    test_data = {}
    for package, pkg_data in file_data.items():
        test_data[package] = ROSCompilePackageFiles(package, pkg_data, executables)
    for pkg_data in config:
        if 'function' in pkg_data:
            pkg_data['functions'] = [pkg_data.pop('function')]

    return config, test_data



from ..core import glinter
from ..util import get_data_file
from ros_introspect.components.rviz_config import dictionary_subtract
import yaml

CLASS_DEFAULTS = yaml.safe_load(get_data_file('rviz_class_defaults.yaml'))
GENERIC_DEFAULTS = yaml.safe_load(get_data_file('rviz_generic_defaults.yaml'))
GLOBAL_DEFAULTS = yaml.safe_load(get_data_file('rviz_global_defaults.yaml'))

ROBOT_MODEL_LINK_DEFAULTS = {'Alpha': 1, 'Show Axes': False, 'Show Trail': False, 'Value': True}


@glinter
def clean_up_rviz_configs(package):
    for rviz_config in package.rviz_config:
        for config in rviz_config.get_class_dicts():
            if dictionary_subtract(config, GENERIC_DEFAULTS):
                rviz_config.changed = True

            full_class = config['Class']
            class_name = full_class.split('/')[-1]

            the_defaults = CLASS_DEFAULTS.get(full_class, {})
            the_defaults.update(CLASS_DEFAULTS.get(class_name, {}))
            if dictionary_subtract(config, the_defaults):
                rviz_config.changed = True

            # Special Case(s)
            if config.get('Topic') == '':
                del config['Topic']
                rviz_config.changed = True

            if full_class == 'rviz_default_plugins/RobotModel':
                for k, v in list(config.get('Links', {}).items()):
                    if not isinstance(v, dict):
                        continue
                    if dictionary_subtract(config['Links'][k], ROBOT_MODEL_LINK_DEFAULTS):
                        rviz_config.changed = True
                        if not config['Links'][k]:
                            del config['Links'][k]

            if full_class in ['rviz/Camera', 'rviz_default_plugins/Camera'] and 'Visibility' in config:
                visibility = config['Visibility']
                for key in list(visibility.keys()):
                    if visibility[key]:
                        rviz_config.changed = True
                        del visibility[key]
                if not visibility:
                    del config['Visibility']

            if full_class in ['rviz/TF', 'rviz_default_plugins/TF']:
                if 'Tree' in config:
                    del config['Tree']
                    rviz_config.changed = True
                if 'Frames' in config:
                    frames = config['Frames']
                    for key in list(frames.keys()):
                        if frames[key]['Value']:
                            del frames[key]
                            rviz_config.changed = True
                    if not frames:
                        del config['Frames']

        if dictionary_subtract(rviz_config.contents, GLOBAL_DEFAULTS):
            rviz_config.changed = True



import argparse
import click
import pathlib
import shutil
import tempfile

from . import get_linters
from .terminal import color_header
from .diff import get_diff_string
from ros_introspect import find_packages, ROSResources, Package
from ros_introspect.finder import walk


# Check if a path is a textfile
# via https://stackoverflow.com/a/7392391/999581
TEXT_CHARS = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})


def is_binary_file(path):
    first_bytes = open(path, 'rb').read(1024)
    return bool(first_bytes.translate(None, TEXT_CHARS))


def copy_text_files(src_folder, dst_folder):
    for subpath in walk(src_folder):
        src_path = src_folder / subpath
        dst_path = dst_folder / subpath
        dst_path.parent.mkdir(exist_ok=True, parents=True)

        if not is_binary_file(src_path):
            shutil.copyfile(src_path, dst_path)
        else:
            # Write stub file
            with open(dst_path, 'wb'):
                pass


def read_or_empty(filepath):
    """Read an existing file or return an empty string"""
    if filepath.exists():
        try:
            return open(filepath).read()
        except UnicodeDecodeError:
            return ''
    else:
        return ''


def get_diff(pkg0, pkg1):
    names0 = set(pkg0.components_by_name.keys())
    names1 = set(pkg1.components_by_name.keys())
    diff = []
    for rel_filename in names0.union(names1):
        contents0 = read_or_empty(pkg0.root / rel_filename)
        contents1 = read_or_empty(pkg1.root / rel_filename)

        if contents0 == contents1:
            continue

        diff.append(get_diff_string(contents0, contents1, rel_filename))

    return diff


def preview_changes(package, fn_name, fne, use_package_name=False):
    """
    Given a package and a single function, run the function on a copy of the package in a temp dir and display the diff.
    """

    try:
        temp_dir = pathlib.Path(tempfile.mkdtemp())
        new_package_root = temp_dir / package.name

        copy_text_files(package.root, new_package_root)
        new_pkg = Package(new_package_root)

        fne(new_pkg)

        if not new_pkg.has_changes():
            return False

        new_pkg.save()

        the_diff = get_diff(package, new_pkg)
        if len(the_diff) == 0:
            return False

        if use_package_name:
            print(color_header(fn_name + ' (' + package.name + ')'))
        else:
            print(color_header(fn_name))

        for diff_string in the_diff:
            print(diff_string)

    finally:
        shutil.rmtree(temp_dir)
    return True


def main():
    linters = get_linters()

    class ValidateCleaner(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            for value in values:
                if value not in linters:
                    raise ValueError(f'{value} not a valid linter!')
            args.linters = values

    parser = argparse.ArgumentParser()
    parser.add_argument('linters', nargs='*', action=ValidateCleaner, metavar='linter', default=[],
                        help='By default, run all linters. If any are specified here, only those specified are run.')
    parser.add_argument('-f', '--folder', type=pathlib.Path, default='.',
                        help='The folder to search for ROS packages in. Defaults to the current directory.')
    parser.add_argument('-y', '--yes-to-all', action='store_true',
                        help='Non-interactive mode that accepts all suggestions.')
    parser.add_argument('-s', '--skip-ros-load', action='store_true',
                        help='Avoid loading ROS resources, useful in scripting environments.')

    args = parser.parse_args()

    pkgs = list(find_packages(args.folder))

    resources = ROSResources.get()

    if not args.skip_ros_load:
        resources.load_from_ros()

    for package in pkgs:
        for name, fne in linters.items():
            if args.linters and name not in args.linters:
                continue
            try:
                if args.yes_to_all:
                    fne(package)
                    package.save()
                    continue

                if preview_changes(package, name, fne, len(pkgs) > 1):
                    if click.confirm('Would you like to make this change?'):
                        fne(package)
                        package.write()
                    print('')

            except click.Abort:
                print('')
                exit(0)
            except Exception as e:
                click.secho(f'Exception occurred while running {name}: {e}', fg='red')
                raise
